#include "MainWindow.h"

MainWindow::MainWindow()
  : _mrd(new MatrixResizeDialog(this))
{
    setupUi(this);
    tableWidget->setRowCount(100);
    tableWidget->setColumnCount(100);

    // capture escape key presses
    installEventFilter(this);

    connect(_mrd, SIGNAL(resize(int, int)), this, SLOT(resized(int, int)));
}

// NOTE: This method will not ask for you to save your work. Normally, this would be
//       added and tested.
bool MainWindow::eventFilter(QObject *obj, QEvent *e)
{
    if(obj == this)
    {
        if(e->type() == QEvent::KeyPress)
        {
            QKeyEvent *key = static_cast<QKeyEvent*>(e);
            if(key->key() == Qt::Key_Escape)
            {
                close();
            }
        }
    }

    return QMainWindow::eventFilter(obj, e);
}

// NOTE: This method does not ask the user to save changes before performing a file load.
//       Normally, this would be added and tested.
//
// NOTE: Loading of a file requires that the user supply the .csv extention instead of the
//       application just supplying it. Out of time to do.
//
// NOTE: As the file loads, it pushes the edits onto the QVector. Unfortunately, this is a
//       space tradeoff for processing speed. I want to be able to have a 20K x 20K matrix
//       be able to save quickly instead of having to iterate over all the cells in the
//       QTableWidget. I used a vector for its quick insertion and linear search since my
//       assumption is that there would not be many cell edits occurring. The use of a vector
//       also has the ability to in the future add a history option. Ctrl + Z, in particular,
//       comes to mind.
//
void MainWindow::on_actionLoad_triggered(bool checked)
{
    disconnect(tableWidget, SIGNAL(cellChanged(int, int)),
               this, SLOT(on_tableWidget_cellChanged(int, int)));

    QFileDialog dialog(this);
    dialog.setNameFilter("*.csv");
    dialog.setAcceptMode(QFileDialog::AcceptOpen);
    QStringList files;
    if(dialog.exec())
    {
        files = dialog.selectedFiles();
        QFile file(files[0]);
        if(not file.open(QIODevice::ReadOnly))
        {
            // Do some error messaging to the StatusBar
            return;
        }

        QTextStream istream(&file);
        // First line should be the size of the previous matrix
        QString line = istream.readLine();
        QStringList items = line.split(',');
        if(items.size() != 2)
        {
            // Funky data; send message to user about failed load
            return;
        }

        bool ok = false;
        int x = 0, y = 0;
        x = items[0].toInt(&ok);
        if(not ok)
        {
            // Funky data; send message to user about failed load
            return;
        }

        y = items[1].toInt(&ok);
        if(not ok)
        {
            // Funky data; send message to user about failed load
            return;
        }

        tableWidget->clear();
        _edits.clear();

        tableWidget->setRowCount(x);
        tableWidget->setColumnCount(y);

        while(not istream.atEnd())
        {
            line = istream.readLine();
            items = line.split(',');
            if(items.size() != 3)
            {
                // Throw an exception here and close the program
                close();
            }

            x = items[0].toInt(&ok);
            if(not ok)
            {
                // Failed to load
                return;
            }
            y = items[1].toInt(&ok);
            if(not ok)
            {
                // Failed to load
                return;
            }
            double val = items[2].toDouble(&ok);
            if(not ok)
            {
                // Failed to load
                return;
            }

            tableWidget->setItem(x, y, new QTableWidgetItem(items[2]));
            _edits.push_back(QPair<int, int>(x, y));
        }
    }

    connect(tableWidget, SIGNAL(cellChanged(int, int)),
            this, SLOT(on_tableWidget_cellChanged(int, int)));
}

void MainWindow::on_actionSave_triggered(bool checked)
{
    QFileDialog dialog(this);
    dialog.setNameFilter("*.csv");
    dialog.setAcceptMode(QFileDialog::AcceptSave);
    QStringList files;
    if(dialog.exec())
    {
        files = dialog.selectedFiles();
        QFile file(files[0]);
        if(not file.open(QIODevice::WriteOnly))
        {
            // Do some error messaging to the StatusBar
            return;
        }

        QTextStream ostream(&file);
        ostream<<tableWidget->rowCount()<<","<<tableWidget->columnCount()<<"\n";
        disconnect(tableWidget, SIGNAL(cellChanged(int, int)),
                   this, SLOT(on_tableWidget_cellChanged(int, int)));
        for(QVector<QPair<int, int>>::iterator i = _edits.begin(); i != _edits.end();
            ++i)
        {
            QPair<int, int> p = *i;
            QTableWidgetItem *item = tableWidget->item(p.first, p.second);
            ostream<<p.first<<","<<p.second<<","<<item->text()<<"\n";
            item->setBackground(QBrush(QColor(Qt::white)));
        }

        connect(tableWidget, SIGNAL(cellChanged(int, int)),
                this, SLOT(on_tableWidget_cellChanged(int, int)));
    }
}

void MainWindow::on_actionQuit_triggered(bool checked)
{
    close();
}

void MainWindow::on_actionSize_triggered(bool checked)
{
    _mrd->exec();
}

void MainWindow::on_tableWidget_cellChanged(int row, int column)
{
    QTableWidgetItem *item = tableWidget->item(row, column);
    QString text = item->text();
    bool ok = false;
    double val = text.toDouble(&ok);
    if(not ok)
    {
        item->setData(Qt::DisplayRole, "");
        return;
    }

    // Look to see if the edited cell is in the _edits list already. If so, do nothing.
    // Otherwise, push the modification onto the _edits list.
    bool found = false;
    for(QVector<QPair<int,int>>::iterator i = _edits.begin(); (i != _edits.end()) and not found; ++i)
    {
        if((i->first == row) and (i->second == column))
        {
            found = true;
        }
    }

    if(not found)
        _edits.push_back(QPair<int, int>(row, column));

    // setBackground causes an emission of the cellChanged signal. The extra call
    // is redundant and not needed; disable the slot connection until after the 
    // background color has changed.
    disconnect(tableWidget, SIGNAL(cellChanged(int, int)),
               this, SLOT(on_tableWidget_cellChanged(int, int)));
    item->setBackground(QBrush(QColor(Qt::yellow)));
    connect(tableWidget, SIGNAL(cellChanged(int, int)),
            this, SLOT(on_tableWidget_cellChanged(int, int)));
}

void MainWindow::resized(int row, int col)
{
    tableWidget->setRowCount(row);
    tableWidget->setColumnCount(col);

    // NOTE: The viewport could have changed sizes to exclude modifications made
    //       to existing cells that have now been deleted. Therefore, this code
    //       would need to iterate through the stack to find elements that have
    //       now been deleted. Then those elements would need removed.
}
